//
///////////////////////////////////////////////////////////////////////////////
//
//    File Name:  clk_adjust_r1.0.v
//      Version:  1.0
//         Date:  22.02.2008
//        Model:  
//
//      Company:  Xilinx, Inc.
//  Contributor:  Martin Kellermann
//
//   Disclaimer:  XILINX IS PROVIDING THIS DESIGN, CODE, OR
//                INFORMATION "AS IS" SOLELY FOR USE IN DEVELOPING
//                PROGRAMS AND SOLUTIONS FOR XILINX DEVICES.  BY
//                PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
//                ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
//                APPLICATION OR STANDARD, XILINX IS MAKING NO
//                REPRESENTATION THAT THIS IMPLEMENTATION IS FREE
//                FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE
//                RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY
//                REQUIRE FOR YOUR IMPLEMENTATION.  XILINX
//                EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH
//                RESPECT TO THE ADEQUACY OF THE IMPLEMENTATION,
//                INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
//                REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
//                FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES
//                OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
//                PURPOSE.
//
//                (c) Copyright 2008 Xilinx, Inc.
//                All rights reserved.
//
///////////////////////////////////////////////////////////////////////////////
// 
// Summary:
//  This module takes the Control-Logic that generates the Faster- and Slower
//  adjustment-commands and monitors those and judges on the state of the 
//  clock-alignment.
//  At startup the DLL is brought near the operating-point and once that is 
//  found the commands are counted and the FIFO-status of the control monitored.
//  Once the Control-logic does not show full- or empty-conditions anymore but
//  only a given number of adjustments, it is assumed that the DLL is in Lock.
//  The Lock-condition is reported to the next higher level of the design-
//  hierarchy.
//--------------------------------------------------------------------------------
`timescale 1ns / 1ps
module clk_adjust_v2 #(
       parameter [7:0] DLL_MULTIPLY = 2)
       (
       input DLL_CLK,
       input REF_CLK,
       input CTRL_CLK,
       input STOP_LOCK_CHECK,
       output DLL_FASTER,
       output DLL_SLOWER,
       output DLL_LOCKED,
       input RESET
       );



parameter [1:0] IDLE = 0;
parameter [1:0] EXACT = 1;
parameter [1:0] LOCKED = 2;

parameter integer WR_ADDR_WIDTH = 4;



reg [1:0] cstate;// = IDLE;
reg [1:0] nstate;// = IDLE;

reg  fifo_rst;
reg  fifo_rst_reg;// =1'b 1;
reg  rd_en;// = 0;

reg  start_wait_fifo;
wire  faster_fifo;
wire  slower_fifo;
wire  full;
wire  empty;
reg  lock;
reg  lock_r;// = 0;
wire  lock_assumed;
//	attribute keep: string;
//	attribute keep of lock_r: signal is "true";
//	attribute fsm_encoding : string;
//	attribute fsm_encoding of cstate: signal is "one-hot";
//	attribute fsm_encoding of nstate: signal is "one-hot";
//	attribute fsm_extract: string;
//	attribute fsm_extract of cstate: signal is "yes";
//	attribute fsm_extract of nstate: signal is "yes";
parameter VCC = 1'b 1;

initial begin
	cstate = IDLE;
	nstate = IDLE;
	fifo_rst_reg = 1'b 1;
	rd_en = 1'b 0;
	lock_r = 1'b 1;
end


  assign DLL_FASTER = faster_fifo;
  assign DLL_SLOWER = slower_fifo;
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  //
  // FIFO Control-Logic (without memory) for Exact Loop (=0 ppm 
  // off target frequency)
  //
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  fifo_ctrl_v2 #(
    .WR_ADDR_WIDTH(WR_ADDR_WIDTH),
    .DLL_MULTIPLY(DLL_MULTIPLY))
  i_fifo_ctrl(
    .WR_CLK(REF_CLK),
    .WR_EN(VCC),
    .RD_CLK(DLL_CLK),
    .RD_EN(rd_en),
    .RST(fifo_rst_reg),
    .FASTER(faster_fifo),
    .SLOWER(slower_fifo),
    .EMPTY(empty),
    .FULL(full),
    .CTRL_CLK(CTRL_CLK));

  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  //
  // Lock-detection 
  //
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  lock_count #(
      .COUNT_LOCK(6))
  i_lock_count(
    .FASTER(faster_fifo),
    .SLOWER(slower_fifo),
    .EMPTY(empty),
    .FULL(full),
    .CTRL_CLK(CTRL_CLK),
    .CHECK_LOCK(rd_en),
    .LOCK(lock_assumed));

  //	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  //	--
  //	-- State-machine for control of frequency-alignment
  //	--
  //	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  always @(posedge CTRL_CLK) begin
    if(RESET == 1'b 1) begin
      cstate <= IDLE;
      lock_r <= 1'b 0;
    end
    else begin
      cstate <= nstate;
      lock_r <= lock;
    end
  end
  assign DLL_LOCKED = lock_r;
  
  
  always @(posedge CTRL_CLK) begin
    start_wait_fifo <= 1'b 0;
    fifo_rst <= 1'b 1;
    case (cstate)
      IDLE: begin
        lock <= 1'b 0;
        fifo_rst <= 1'b 1;
        nstate <= EXACT;
      end
      EXACT: begin
        fifo_rst <= 1'b 0;
        start_wait_fifo <= 1'b 1;
        lock <= 1'b 0;
        if (lock_assumed == 1'b 1)
          nstate <= LOCKED;
        else
          nstate <= EXACT;
      end
      LOCKED: begin
        lock <= 1'b 1;
        fifo_rst<= 1'b 0;

	if (STOP_LOCK_CHECK == 1'b 1)
	   nstate <= LOCKED;
        else begin
	   if ((empty == 1'b 1) || (full == 1'b 1))
	     nstate <= EXACT;
           else
             nstate <= LOCKED;
        end

      end
      default:begin
        nstate <= IDLE;
      end
    endcase
  end


  always @(posedge DLL_CLK) begin
    if(RESET == 1'b 1) begin
      fifo_rst_reg <= 1'b 1;
    end
    else begin
      fifo_rst_reg <= fifo_rst;
    end
  end

  always @(posedge DLL_CLK) begin
    if((RESET == 1'b 1)) begin
      rd_en <= 1'b 0;
    end
    else if(start_wait_fifo == 1'b 1) begin
      rd_en <= 1'b 1;
    end
  end


endmodule
