//
///////////////////////////////////////////////////////////////////////////////
//
//    File Name:  fifo_ctrl_r1.1.v
//      Version:  1.1
//         Date:  22.02.2008
//        Model:
//
//      Company:  Xilinx, Inc.
//  Contributor:  Martin Kellermann
//
//   Disclaimer:  XILINX IS PROVIDING THIS DESIGN, CODE, OR
//                INFORMATION "AS IS" SOLELY FOR USE IN DEVELOPING
//                PROGRAMS AND SOLUTIONS FOR XILINX DEVICES.  BY
//                PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
//                ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
//                APPLICATION OR STANDARD, XILINX IS MAKING NO
//                REPRESENTATION THAT THIS IMPLEMENTATION IS FREE
//                FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE
//                RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY
//                REQUIRE FOR YOUR IMPLEMENTATION.  XILINX
//                EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH
//                RESPECT TO THE ADEQUACY OF THE IMPLEMENTATION,
//                INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
//                REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
//                FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES
//                OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
//                PURPOSE.
//
//                (c) Copyright 2008 Xilinx, Inc.
//                All rights reserved.
//
///////////////////////////////////////////////////////////////////////////////
// 
// Summary:
//  This module is responsible to control the faster- and slower-commands to the 
//  oscillator, based on the percentage how full the "FIFO" is. 
//  The control is based on the structure of a FIFO, however no memory is used. 
//  If the FIFO-Ctrl detects, that it is filling up to much then the command to
//  the oscillator is issued to speed up. If the FIFO is moving towards empty then
//  the oscillator receives the command to slow down.
//  When the DLL is locked typically only the middle-positions of level_int are
//  active.
//--------------------------------------------------------------------------------

`timescale 1ns / 1ps

module fifo_ctrl_v2#(
	parameter [3:0] WR_ADDR_WIDTH = 4,
	parameter [7:0] DLL_MULTIPLY = 2)
	(
	input WR_CLK,
	input WR_EN,
	input RD_CLK,
	input RD_EN,
	input RST,
	output FASTER,
	output SLOWER,
	input CTRL_CLK,
	output FULL,
	output EMPTY
	);

	`include "./iodll_src/pkg_dll_reco_r1.0.v"

	localparam integer max_size_wr = 2 ** (WR_ADDR_WIDTH + 1) - 1;
	localparam integer max_size_rd = ((max_size_wr + 1) * DLL_MULTIPLY)-1;
	localparam real max_size_rd_real = max_size_rd;

	localparam integer RD_ADDR_WIDTH = round_up_func(log_bin_func(max_size_rd_real)) ;
	localparam integer RD_ADDR_HALF = max_size_rd / 4;
	localparam integer WR_ADDR_HALF = RD_ADDR_HALF;
	localparam integer WR_ADDR_FULL = RD_ADDR_HALF * 2;

	localparam [RD_ADDR_WIDTH:0] WR_ADDR_HALF_SL = WR_ADDR_HALF[RD_ADDR_WIDTH:0];

	localparam integer OUTER_FAST_BOUNDARY = max_size_rd * 3 / 16;
	localparam integer INNER_FAST_BOUNDARY = max_size_rd * 6 / 16;
	localparam integer NEUTRAL = max_size_rd / 2;
	localparam integer INNER_SLOW_BOUNDARY = max_size_rd - INNER_FAST_BOUNDARY - 1;
	localparam integer OUTER_SLOW_BOUNDARY = max_size_rd - OUTER_FAST_BOUNDARY - 1;

	reg [RD_ADDR_WIDTH:0] wr_addr; 
	wire [RD_ADDR_WIDTH:0] wr_addr_rd_c;
	reg [RD_ADDR_WIDTH:0] wr_addr_rd;

	reg [RD_ADDR_WIDTH:0] rd_addr;// :=(others => '0');

	localparam [RD_ADDR_WIDTH:0] WR_ADDR_GRAY_RESET = {1'b 0, WR_ADDR_HALF_SL [RD_ADDR_WIDTH : 1]} ^ WR_ADDR_HALF_SL ;
	reg [RD_ADDR_WIDTH:0] wr_addr_gray;// := WR_ADDR_GRAY_RESET;
	reg [RD_ADDR_WIDTH:0] wr_addr_gray_rd;// := WR_ADDR_GRAY_RESET;

	wire [RD_ADDR_WIDTH:0] wr_addr_c;
	wire [RD_ADDR_WIDTH:0] rd_addr_c;
	reg [RD_ADDR_WIDTH:0] wr_addr_12_16;
	reg [RD_ADDR_WIDTH:0] wr_addr_9_16;
	reg [RD_ADDR_WIDTH:0] wr_addr_1_2;
	reg [RD_ADDR_WIDTH:0] wr_addr_6_16;
	reg [RD_ADDR_WIDTH:0] wr_addr_3_16;

	reg  full_int;
	reg  empty_int;
	reg  full_set;
	reg  empty_set;

	reg  full_ctlr;
	reg  empty_ctlr;

	reg faster_int;// = 0;
	reg slower_int;// = 0;
	assign FASTER = faster_int;
	assign SLOWER = slower_int;


	reg [3:0] diff_int;
	reg [3:0] diff_prev;

	reg [4:0] level_int;//:="00111";
	reg [4:0] level_int_prev;//:="00111";
	//level_int = 00001:  3/16 full
	//level_int = 00011:  6/16 full
	//level_int = 00111:  1/2
	//level_int = 01111:  10/16
	//level_int = 11111:  13/16 full

	reg [4:0] active_level;//:="00111";
	reg level_toggle;

	reg [1:0] tendency;
	reg [1:0] tendency_prev;
	reg [1:0] tendency_prev2;
	reg [1:0] resulting_tendency;//:=(others => '0');

	reg  faster_rdclk;
	reg  slower_rdclk;// := '0';
	reg [2:0] faster_vec;
	reg [2:0] slower_vec;//:="000";

	wire  set_faster;
	wire  set_slower;
	reg  clear_faster;
	reg  clear_slower;//:='0';

	localparam integer max_command_cnt = 23 * DLL_MULTIPLY;
	reg  command;//:='0';
	integer command_cnt;// := 0;
	reg  command_cnt_tc;// :='0';
	reg  up_coarse;
	reg  down_coarse;//:='0';
	reg  coarse_done;//:='0';

//--------------------
	localparam integer max_cnt_coarse_wr = round_up_func(100 / DLL_MULTIPLY);
	localparam integer max_cnt_coarse_wr_width = round_up_func(log_bin_func(max_cnt_coarse_wr))  + 1;
	localparam integer max_cnt_coarse_rd_width = round_up_func(log_bin_func(DLL_MULTIPLY))  + max_cnt_coarse_wr_width;

	reg [max_cnt_coarse_wr_width:0] cnt_wr;//:=(others=>'0');--driven by Refclk
	reg [max_cnt_coarse_rd_width:0] cnt_rd;//:=(others=>'0');--driven by DLL-clock

	reg  rst_cnt_rd;//:='0';--\ synchronous reset for the two
	reg  rst_cnt_wr;//:='0';--/ counters

	reg  cnt_rd_run;//:='0';--\ enable-signals for the two
	reg  cnt_wr_run;//:='0';--/ counters
	reg  cnt_rd_stopped;//:='0';
//===========================

	initial begin
		wr_addr = WR_ADDR_HALF;
		wr_addr_rd = WR_ADDR_HALF;
		rd_addr = 0;
		wr_addr_gray = WR_ADDR_GRAY_RESET;
		wr_addr_gray_rd = WR_ADDR_GRAY_RESET;
		full_int = 0;
		empty_int = 0;
		full_set = 0;
		empty_set = 0;
		full_ctlr = 0;
		empty_ctlr = 0;
		diff_int = 0;
		diff_prev = 0;
		level_int = 5'b 00111;
		level_int_prev = 5'b 00111;
		active_level = 5'b 00111;
		tendency = 0;
		tendency_prev = 0;
		tendency_prev2 = 0;
		resulting_tendency = 0;
		faster_rdclk = 0;
		slower_rdclk = 0;
		faster_vec = 0;
		slower_vec = 0;
		faster_int = 0;
		slower_int = 0;
		clear_faster = 0;
		clear_slower = 0;
		command = 0;
		command_cnt_tc = 0;
		up_coarse = 0;
		down_coarse = 0;
		coarse_done = 0;
		cnt_wr = 0;
		cnt_rd = 0;
		rst_cnt_rd = 0;
		rst_cnt_wr = 0;
		cnt_rd_run = 0;
		cnt_wr_run = 0;
		cnt_rd_stopped = 0;
	end


	wire [15:0] debug_dist;

	//synthesis translate off
	//Display of calculated constants in Simulator
	initial begin
		//parameter string
		$display("");
		$display("");
		$display("Calculated Constants");
		$display("~~~~~~~~~~~~~~~~~~~~");
		$display("max_cnt_coarse_wr: %d", max_cnt_coarse_wr);
		$display("max_cnt_coarse_wr_width: %d", max_cnt_coarse_wr_width);
		$display("max_cnt_coarse_rd_width: %d", max_cnt_coarse_rd_width);
		$display("max_size_wr: %d", max_size_wr);
		$display("max_size_rd: %d", max_size_rd);
		$display("RD_ADDR_WIDTH: %d", RD_ADDR_WIDTH);
		$display("RD_ADDR_HALF: %d", RD_ADDR_HALF);
		$display("WR_ADDR_HALF: %d", WR_ADDR_HALF);
		$display("WR_ADDR_FULL: %d", WR_ADDR_FULL);
		$display("WR_ADDR_GRAY_RESET: %d", WR_ADDR_GRAY_RESET);
		$display("OUTER_FAST_BOUNDARY: %d", OUTER_FAST_BOUNDARY);
		$display("INNER_FAST_BOUNDARY: %d", INNER_FAST_BOUNDARY);
		$display("INNER_SLOW_BOUNDARY: %d", INNER_SLOW_BOUNDARY);
		$display("OUTER_SLOW_BOUNDARY: %d", OUTER_SLOW_BOUNDARY);
		//$sdisplay(": %d", );

	end
	//synthesis translate on

  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  //
  // Write-Counter and coding to Gray for clock-transfer
  //
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  always @(posedge WR_CLK) begin
    if(RST == 1'b 1 || coarse_done == 1'b 0) begin
      wr_addr <= WR_ADDR_HALF_SL;
      wr_addr_gray <= WR_ADDR_GRAY_RESET;
    end
    else if(WR_EN == 1'b 1) begin
      wr_addr <= wr_addr_c;
      wr_addr_gray <= ({1'b 0,wr_addr_c[RD_ADDR_WIDTH:1] }) ^ wr_addr_c;
    end
  end

  assign wr_addr_c = wr_addr + DLL_MULTIPLY;
	
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  //
  // Read-Counter
  //
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  always @(posedge RD_CLK) begin
    if(RST == 1'b 1 || coarse_done == 1'b 0) begin
      rd_addr <= 0;
    end
    else if(RD_EN == 1'b 1) begin
      rd_addr <= rd_addr_c;
    end
  end

  assign rd_addr_c = rd_addr + 1'b 1;
	
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  //
  // Clk-Domain transfer and re-coding to binary
  //
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  always @(posedge RD_CLK) begin
    if(RST == 1'b 1) begin
      wr_addr_gray_rd <= WR_ADDR_GRAY_RESET;
      wr_addr_rd <= WR_ADDR_HALF_SL;
      // Level-counters:
      wr_addr_12_16 <= WR_ADDR_HALF_SL + OUTER_SLOW_BOUNDARY;
      wr_addr_9_16 <= WR_ADDR_HALF_SL + INNER_SLOW_BOUNDARY;
      wr_addr_1_2 <= WR_ADDR_HALF_SL + NEUTRAL;
      wr_addr_6_16 <= WR_ADDR_HALF_SL + INNER_FAST_BOUNDARY;
      wr_addr_3_16 <= WR_ADDR_HALF_SL + OUTER_SLOW_BOUNDARY;
    end
    else begin
      wr_addr_gray_rd <= wr_addr_gray;
      wr_addr_rd <= wr_addr_rd_c;
      //Level counters:
      wr_addr_12_16 <= wr_addr_rd_c + OUTER_SLOW_BOUNDARY;
      wr_addr_9_16 <= wr_addr_rd_c + INNER_SLOW_BOUNDARY;
      wr_addr_1_2 <= wr_addr_rd_c + NEUTRAL;
      wr_addr_6_16 <= wr_addr_rd_c + INNER_FAST_BOUNDARY;
      wr_addr_3_16 <= wr_addr_rd_c + OUTER_FAST_BOUNDARY;
    end
  end

  assign wr_addr_rd_c = wr_addr_gray_rd ^ ({1'b 0,wr_addr_rd_c[RD_ADDR_WIDTH:1] });
  assign debug_dist = (wr_addr_rd  > rd_addr) ? (wr_addr_rd - rd_addr) : max_size_rd - (rd_addr - wr_addr_rd);
	
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  //
  // Generation of Full- and Empty-Flag
  //
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  always @(posedge RD_CLK) begin
    if(RST == 1'b 1) begin
      full_int <= 1'b 0;
      empty_int <= 1'b 0;
    end
    else begin
      if(wr_addr_rd == rd_addr) begin
        empty_int <= 1'b 1;
      end
      else begin
        empty_int <= 1'b 0;
      end
      if((wr_addr_rd[RD_ADDR_WIDTH]  != rd_addr[RD_ADDR_WIDTH] ) && (wr_addr_rd[RD_ADDR_WIDTH - 1]  != rd_addr[RD_ADDR_WIDTH - 1] ) && (wr_addr_rd[RD_ADDR_WIDTH - 2:0]  == rd_addr[RD_ADDR_WIDTH - 2:0] )) begin
        full_int <= 1'b 1;
      end
      else begin
        full_int <= 1'b 0;
      end
    end
  end

  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  //
  // Transfer of Full- and Empty-Flag into CTRL_CLK-domain
  //
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  always @(posedge RD_CLK) begin
    if(RST == 1'b 1) begin
      full_set <= 1'b 0;
      empty_set <= 1'b 0;
    end
    else begin
      if(full_int == 1'b 1) begin
        full_set <= 1'b 1;
      end
      else if(full_ctlr == 1'b 1) begin
        full_set <= 1'b 0;
      end
      if(empty_int == 1'b 1) begin
        empty_set <= 1'b 1;
      end
      else if(empty_ctlr == 1'b 1) begin
        empty_set <= 1'b 0;
      end
    end
  end

  always @(posedge CTRL_CLK) begin
    full_ctlr <= full_set;
    empty_ctlr <= empty_set;
  end

  assign FULL = full_ctlr;
  assign EMPTY = empty_ctlr;
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  //
  // Check on level of FIFO-filling
  //
  //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  always @(posedge RD_CLK) begin
    if(RST == 1'b 1) begin
      level_int <= 5'b 00111;
      level_int_prev <= 5'b 00111;
    end
    else begin
      level_int_prev <= level_int;
      //12/16-FULL: (=outer slow boundary)
      if (((wr_addr_12_16[RD_ADDR_WIDTH - 1]  != rd_addr[RD_ADDR_WIDTH - 1] ) &&
			   (wr_addr_12_16[RD_ADDR_WIDTH - 2:0]  >= rd_addr[RD_ADDR_WIDTH - 2:0] )) ||
				 ((wr_addr_12_16[RD_ADDR_WIDTH - 1]  == rd_addr[RD_ADDR_WIDTH - 1] ) &&
				 (wr_addr_12_16[RD_ADDR_WIDTH - 2:0]  < rd_addr[RD_ADDR_WIDTH - 2:0] ))) begin
        level_int[4]  <= 1'b 1;
      end
      else begin
        level_int[4]  <= 1'b 0;
      end
      //9/16-FULL: (=inner slow boundary)
      if(((wr_addr_9_16[RD_ADDR_WIDTH - 1]  != rd_addr[RD_ADDR_WIDTH - 1] ) &&
			   (wr_addr_9_16[RD_ADDR_WIDTH - 2:0]  >= rd_addr[RD_ADDR_WIDTH - 2:0] )) ||
				 ((wr_addr_9_16[RD_ADDR_WIDTH - 1]  == rd_addr[RD_ADDR_WIDTH - 1] ) &&
				 (wr_addr_9_16[RD_ADDR_WIDTH - 2:0]  < rd_addr[RD_ADDR_WIDTH - 2:0] ))) begin
        level_int[3]  <= 1'b 1;
      end
      else begin
        level_int[3]  <= 1'b 0;
      end
      //HALF-FULL: (=neutral area)
      if(((wr_addr_1_2[RD_ADDR_WIDTH - 1]  != rd_addr[RD_ADDR_WIDTH - 1] ) &&
			   (wr_addr_1_2[RD_ADDR_WIDTH - 2:0]  >= rd_addr[RD_ADDR_WIDTH - 2:0] )) ||
				 ((wr_addr_1_2[RD_ADDR_WIDTH - 1]  == rd_addr[RD_ADDR_WIDTH - 1] ) &&
				 (wr_addr_1_2[RD_ADDR_WIDTH - 2:0]  < rd_addr[RD_ADDR_WIDTH - 2:0] ))) begin
        level_int[2]  <= 1'b 1;
      end
      else begin
        level_int[2]  <= 1'b 0;
      end
      //6/16 FULL (=inner fast boundary)
      if(((wr_addr_6_16[RD_ADDR_WIDTH - 1]  != rd_addr[RD_ADDR_WIDTH - 1] ) &&
			   (wr_addr_6_16[RD_ADDR_WIDTH - 2:0]  >= rd_addr[RD_ADDR_WIDTH - 2:0] )) ||
				 ((wr_addr_6_16[RD_ADDR_WIDTH - 1]  == rd_addr[RD_ADDR_WIDTH - 1] ) &&
				 (wr_addr_6_16[RD_ADDR_WIDTH - 2:0]  < rd_addr[RD_ADDR_WIDTH - 2:0] ))) begin
        // or
        level_int[1]  <= 1'b 1;
      end
      else begin
        level_int[1]  <= 1'b 0;
      end
      //3/16 FULL (=outer fast boundary)
      if(((wr_addr_3_16[RD_ADDR_WIDTH - 1]  != rd_addr[RD_ADDR_WIDTH - 1] ) &&
			   (wr_addr_3_16[RD_ADDR_WIDTH - 2:0]  >= rd_addr[RD_ADDR_WIDTH - 2:0] )) ||
				 ((wr_addr_3_16[RD_ADDR_WIDTH - 1]  == rd_addr[RD_ADDR_WIDTH - 1] ) &&
				 (wr_addr_3_16[RD_ADDR_WIDTH - 2:0]  < rd_addr[RD_ADDR_WIDTH - 2:0] ))) begin
        level_int[0]  <= 1'b 1;
      end
      else begin
        level_int[0]  <= 1'b 0;
      end
    end
  end

	always @(level_int or level_int_prev)
	begin
		if(level_int_prev != level_int) 
		begin
			level_toggle <= 1'b 1;
		end
		else begin
			level_toggle <= 1'b 0;
		end
	end

	integer i;

	always @(posedge RD_CLK) begin
		//if(level_int_prev != level_int) begin
		if(level_toggle == 1'b 1) begin
			tendency <= 2'b 00;
			for (i=0; i<=3; i=i+1)
			begin: slow_check
				if((level_int[i]  == 1'b 1 && level_int_prev[i]  == 1'b 0 && level_int[i + 1]  == 1'b 1 && level_int_prev[i + 1]  == 1'b 1) || (level_int[i]  == 1'b 1 && level_int_prev[i]  == 1'b 1 && level_int[i + 1]  == 1'b 0 && level_int_prev[i + 1]  == 1'b 1)) begin
					tendency[1]  <= 1'b 1;
				end
			end
			for (i=0; i<=3; i=i+1)
			begin: fast_check
				if((level_int[i]  == 1'b 1 && level_int_prev[i]  == 1'b 1 && level_int[i+1]  == 1'b 1 && level_int_prev[i+1]  == 1'b 0) || (level_int[i]  == 1'b 0 && level_int_prev[i]  == 1'b 0 && level_int[i+1]  == 1'b 0 && level_int_prev[i+1]  == 1'b 1) || (level_int[i]  == 1'b 0 && level_int_prev[i]  == 1'b 1 && level_int[i+1]  == 1'b 1 && level_int_prev[i+1]  == 1'b 1)) begin
					tendency[0]  <= 1'b 1;
				end
			end
		end
		tendency_prev <= tendency;
		tendency_prev2 <= tendency_prev;
		resulting_tendency <= tendency & tendency_prev & tendency_prev2;
	end

  //Several situations can occur:
  // 1) LOCKED-sitation: FIFO moving around the half-full point
  // 2) RD-clock faster than WR-clock: => FIFO underflowing (during locking-phase)
  // 3) RD-clock slower than WR-clock: => FIFO overflowing (during locking-phase)
  //    The FIFO is started with nominally half-full FIFO-pointers. If nothing would 
  //    be compensated (=RD-clock always slower) then first the FULL-flag is asserted 
  //    and the FIFO-pointers roll over. Later on, with the RD-clock assumed constant
  //    the EMPTY-flag is asserted and again the pointers start catching up leading 
  //    to a FULL-condition (and so on).
  //
  //    If it is assumed that during the locking-phase over- and underflows are 
  //    allowed then care must be taken to distinguish between the faster and the
  //    slower read-situation (with the write-clock being constant).
  //    A (constantly) slower read-clock creates a FULL-flag, overflows and then goes 
  //    on until until it is overtaken by the write-pointer, leading to an EMPTY-flag
  //    being asserted. 
  //    Even though after LOCK both over- or underflowing are not allowed, they can
  //    be used to get into the stable middle-condition:
  //    If a overflow has happended, it is known that the readpointer was to slow, the
  //    speed of reading needs to be increased.
  //    As the symptoms for underflow (= (Empty = 1) & (level_int = 11000)) or overflow
  //    (= (Full=1) & (level_int = 00111)) are the same for both a faster and a slower
  //    read-clock, it is necessary to have a tendency-information. 
  //    If the 1s in level_int are shifting from 4=>0 then the read-clock is too slow, 
  //    on a shift of 1s from 0=>4 the read-clock is too fast. As the tendancy is only
  //    happening relatively slowly, it can only be calculated on changes of the
  //    level_int-signal.
  //    For the overflow/underflow this means that the full/empty-signal is taken and 
  //    checked with the tendency-signal. If the full/empty-signal is asserted with a 
  //    tendency of a read-clock being faster than the write-clock then the tap-number 
  //    must be increased => DLL needs to be slower.
  //    If the full/empty-signal is asserted with a tendency of the read-clock being 
  //    slower than the write-clock, then the read-clock needs to be speeded up by 
  //    decreasing the number of taps in the DLL.
  always @(posedge RD_CLK) begin
    if(RST == 1'b 1) begin
      up_coarse <= 1'b 0;
      down_coarse <= 1'b 0;
      coarse_done <= 1'b 0;
    end
    else begin
      if(slower_rdclk == 1'b 1) begin
        down_coarse <= 1'b 1;
      end
      if(faster_rdclk == 1'b 1) begin
        up_coarse <= 1'b 1;
      end
      if(up_coarse == 1'b 1 && down_coarse == 1'b 1) begin
        coarse_done <= 1'b 1;
      end
    end
  end

  // Following Level_int-combination have a proper meaning:
  // Level_int(4..0)   FIFO-filling:
  //   11000              very close to empty => go slower
  //   11100              close to empty => too fast => go slower
  //   11110              just below half-full => too fast => go slower
  //   01110              half-full (about 1/4th in the middle) ==> wanted situation
  //   01111              just above half-full => too slow => go faster
  //   00111              close to full => too slow => go faster   
  //   00011              very close to full, go faster   	
  always @(posedge RD_CLK) begin : P5
    reg [4:0] level_int_v;

    if(RST == 1'b 1) begin
      faster_rdclk <= 1'b 0;
      slower_rdclk <= 1'b 0;
      active_level <= 5'b 11100;
    end
    else if(RD_EN == 1'b 1) begin
      faster_rdclk <= 1'b 0;
      slower_rdclk <= 1'b 0;
      //two situations need to be considered:
      // a) before lock, over/underflows can occur
      // b) after lock, FIFO is moving within its valid range
      //
      if(level_int != level_int_prev) begin
        level_int_v = level_int;
      end
      else begin
        level_int_v = 0;
      end
      if(command == 1'b 0) begin
        if(coarse_done == 1'b 0) begin
          //Two counters running against each other
          if(cnt_rd_stopped == 1'b 1 && rst_cnt_rd == 1'b 0) begin
            if(cnt_rd > (cnt_wr * DLL_MULTIPLY)) begin
              slower_rdclk <= 1'b 1;
            end
            else if(cnt_wr != 0) begin
              faster_rdclk <= 1'b 1;
            end
          end
          if(cnt_rd_stopped == 1'b 1) begin
            //handshaking for restart
            rst_cnt_rd <= 1'b 1;
          end
          else begin
            rst_cnt_rd <= 1'b 0;
          end
        end
        else begin
          // b) Check within the valid range after lock
          //
          // Only level-changes that are coming from the inner range have an effect on 
          // the number of taps. 
          case(level_int_v)
          5'b 00000 : begin
          end
          5'b 11000 : begin
            active_level <= 5'b 11000;
            if(active_level != 5'b 11000) begin
              slower_rdclk <= 1'b 1;
            end
          end
          5'b 11100 : begin
            active_level <= 5'b 11100;
            if((active_level != 5'b 11100) && (active_level != 5'b 11000)) begin
              slower_rdclk <= 1'b 1;
            end
          end
          5'b 11110 : begin
            active_level <= 5'b 11110;
            if((active_level != 5'b 11110) && (active_level != 5'b 11100) && (active_level != 5'b 11000)) begin
              slower_rdclk <= 1'b 1;
            end
          end
          5'b 01110 : begin
          end
          5'b 01111 : begin
            active_level <= 5'b 01111;
            if((active_level != 5'b 01111) && (active_level != 5'b 00111) && (active_level != 5'b 00011)) begin
              faster_rdclk <= 1'b 1;
            end
          end
          5'b 00111 : begin
            active_level <= 5'b 00111;
            if((active_level != 5'b 00111) && (active_level != 5'b 00011)) begin
              faster_rdclk <= 1'b 1;
            end
          end
          5'b 00011 : begin
            active_level <= 5'b 00011;
            if(active_level != 5'b 00011) begin
              faster_rdclk <= 1'b 1;
            end
          end
          default : begin
          end
          endcase
        end
      end
    end
  end

  //##################################################
  // 
  // Coarse-Loop Counters
  //
  //##################################################

  always @(posedge RD_CLK) begin : P4 // = DLL-Clock

    if(rst_cnt_rd == 1'b 1) begin
      cnt_rd <= 0;
    end
    else if(coarse_done == 1'b 0 && cnt_rd_run == 1'b 1) begin
      cnt_rd <= cnt_rd + 1;
    end
    cnt_rd_stopped <=  ~cnt_rd_run;
  end

  always @(posedge RD_CLK) begin : P3 // = DLL-Clock

    cnt_rd_run <= cnt_wr_run;
    //resynchronisation to Rd_CLK
  end

  always @(posedge WR_CLK) begin : P2 // = Ref-Clock
    rst_cnt_wr <= rst_cnt_rd;
  end

  always @(posedge WR_CLK) begin : P1 // = Ref-Clock
    if(rst_cnt_wr == 1'b 1) begin
      cnt_wr <= 0;
    end
    else if(coarse_done == 1'b 0 && cnt_wr_run == 1'b 1) begin
      cnt_wr <= cnt_wr + 1;
    end
    if(cnt_wr == max_cnt_coarse_wr) begin
      cnt_wr_run <= 1'b 0;
    end
    else if(rst_cnt_wr == 1'b 1) begin
      cnt_wr_run <= 1'b 1;
    end
  end

  //##################################################
  //
  // Damping to avoid too fast switching
  //
  //##################################################
  always @(posedge RD_CLK) begin
    if(RST == 1'b 1) begin
      command <= 1'b 0;
    end
    else if((coarse_done == 1'b 1) && (faster_rdclk == 1'b 1 || slower_rdclk == 1'b 1)) begin
      command <= 1'b 1;
    end
    else if((command_cnt_tc == 1'b 1) && (level_toggle == 1'b 0)) begin
      command <= 1'b 0;
    end
  end

  always @(posedge CTRL_CLK) begin
    if(RST == 1'b 1) begin
      command_cnt <= 0;
    end
    else if(command == 1'b 1) begin
      command_cnt <= command_cnt + 1;
      if(command_cnt >= (max_command_cnt - 3)) begin
        command_cnt_tc <= 1'b 1;
      end
      else begin
        command_cnt_tc <= 1'b 0;
      end
    end
    else begin
      command_cnt <= 0;
      command_cnt_tc <= 1'b 0;
    end
  end

  //##################################################
  // 
  // Translation into CTRLCLK-domain
  //
  //##################################################
  FDC faster_ff(
    .D(1'b 1),
    .C(faster_rdclk),
    .CLR(clear_faster),
    .Q(set_faster));

  FDC slower_ff(
      .D(1'b 1),
    .C(slower_rdclk),
    .CLR(clear_slower),
    .Q(set_slower));

  always @ (set_faster)
	begin
		faster_vec[2] <= set_faster;
	end
	always @ (set_slower)
	begin
		slower_vec[2] <= set_slower;
	end

  always @(posedge CTRL_CLK) begin
    if(RST == 1'b 1) begin
      faster_vec[1:0]  <= 2'b 00;
      slower_vec[1:0]  <= 2'b 00;
      faster_int <= 1'b 0;
      slower_int <= 1'b 0;
      clear_faster <= 1'b 0;
      clear_slower <= 1'b 0;
    end
    else begin
      if(RD_EN == 1'b 1) begin
        faster_vec[1:0]  <= faster_vec[2:1] ;
        slower_vec[1:0]  <= slower_vec[2:1] ;
        if(faster_vec == 3'b 110) begin
          clear_faster <= 1'b 1;
          faster_int <= 1'b 1;
        end
        else begin
          clear_faster <= 1'b 0;
          faster_int <= 1'b 0;
        end
        if(slower_vec == 3'b 110) begin
          slower_int <= 1'b 1;
          clear_slower <= 1'b 1;
        end
        else begin
          clear_slower <= 1'b 0;
          slower_int <= 1'b 0;
        end
      end
      else begin
        faster_int <= 1'b 0;
        slower_int <= 1'b 0;
      end
    end
  end


endmodule
