--
--/////////////////////////////////////////////////////////////////////////////
--
--    File Name:  clk_adjust_r1.0.vhd
--      Version:  1.0
--         Date:  22.02.2008
--        Model:  
--
--      Company:  Xilinx, Inc.
--  Contributor:  
--
--   Disclaimer:  XILINX IS PROVIDING THIS DESIGN, CODE, OR
--                INFORMATION "AS IS" SOLELY FOR USE IN DEVELOPING
--                PROGRAMS AND SOLUTIONS FOR XILINX DEVICES.  BY
--                PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
--                ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
--                APPLICATION OR STANDARD, XILINX IS MAKING NO
--                REPRESENTATION THAT THIS IMPLEMENTATION IS FREE
--                FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE
--                RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY
--                REQUIRE FOR YOUR IMPLEMENTATION.  XILINX
--                EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH
--                RESPECT TO THE ADEQUACY OF THE IMPLEMENTATION,
--                INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
--                REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
--                FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES
--                OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
--                PURPOSE.
--
--                (c) Copyright 2008 Xilinx, Inc.
--                All rights reserved.
--
--/////////////////////////////////////////////////////////////////////////////
-- 
-- Summary:
--  This module takes the Control-Logic that generates the Faster- and Slower
--  adjustment-commands and monitors those and judges on the state of the 
--  clock-alignment.
--  At startup the DLL is brought near the operating-point and once that is 
--  found the commands are counted and the FIFO-status of the control monitored.
--  Once the Control-logic does not show full- or empty-conditions anymore but
--  only a given number of adjustments, it is assumed that the DLL is in Lock.
--  The Lock-condition is reported to the next higher level of the design-
--  hierarchy.
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library work;
use work.pkg_dll_reco.all;

entity clk_adjust_v2 is
	Generic(
		DLL_MULTIPLY: integer range 1 to 256:=2
	);
		Port ( DLL_CLK : in  STD_LOGIC;
					 REF_CLK : in  STD_LOGIC;
					 CTRL_CLK : in  STD_LOGIC;
					 STOP_LOCK_CHECK : in  STD_LOGIC;
					 DLL_FASTER : out  STD_LOGIC:='0';
					 DLL_SLOWER : out  STD_LOGIC:='0';
					 DLL_LOCKED : out STD_LOGIC:='0';
					 RESET : in  STD_LOGIC);
end clk_adjust_v2;

architecture Behavioral of clk_adjust_v2 is
	COMPONENT fifo_ctrl_v2
	GENERIC( 
		WR_ADDR_WIDTH: integer range 1 to 15:=4;
		DLL_MULTIPLY: integer range 1 to 256:=2
	);
	PORT(
		WR_CLK : IN std_logic;
		WR_EN : IN std_logic;
		RD_CLK : IN std_logic;
		RD_EN : IN std_logic;
		RST : IN std_logic;     
		FASTER : out  STD_LOGIC:='0';
		SLOWER : out  STD_LOGIC:='0';
		EMPTY : out  STD_LOGIC:='0'; --created with CTRL_CLK
		FULL : out  STD_LOGIC:='0';  --created with CTRL_CLK
		CTRL_CLK : in  STD_LOGIC);	
	END COMPONENT;

	COMPONENT lock_count
	GENERIC(COUNT_LOCK: integer range 4 to 64:=16);
	PORT(
		FASTER : IN std_logic;
		SLOWER : IN std_logic;
		EMPTY : IN std_logic;
		FULL : IN std_logic;
		CTRL_CLK : IN std_logic;
		CHECK_LOCK : IN std_logic;
		LOCK : OUT std_logic
		);
	END COMPONENT;

	constant WR_ADDR_WIDTH : integer :=4;

	type dll_state is (IDLE, EXACT, LOCKED);
	signal cstate, nstate: dll_state := IDLE;

	signal fifo_rst: std_logic;
	signal fifo_rst_reg: std_logic:='1';
	signal rd_en: std_logic:='0';
	signal start_wait_fifo: std_logic;

	signal faster_fifo, slower_fifo: std_logic;


	signal full, empty: std_logic;
	signal lock: std_logic;
	signal lock_r: std_logic:='0';
	signal lock_assumed: std_logic;

	attribute keep: string;
	attribute keep of lock_r: signal is "true";

	attribute fsm_encoding : string;
	attribute fsm_encoding of cstate: signal is "one-hot";
	attribute fsm_encoding of nstate: signal is "one-hot";
	attribute fsm_extract: string;
	attribute fsm_extract of cstate: signal is "yes";
	attribute fsm_extract of nstate: signal is "yes";

begin

	DLL_FASTER <= faster_fifo;
	DLL_SLOWER <= slower_fifo;

	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	--
	-- FIFO Control-Logic (without memory) for Exact Loop (=0 ppm 
	-- off target frequency)
	--
	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

	i_fifo_ctrl: fifo_ctrl_v2
	GENERIC MAP(
		WR_ADDR_WIDTH => WR_ADDR_WIDTH,
		DLL_MULTIPLY => DLL_MULTIPLY
	)
	PORT MAP(
		WR_CLK => REF_CLK,
		WR_EN => '1',
		RD_CLK => DLL_CLK,
		RD_EN => rd_en,
		RST => fifo_rst_reg,
		FASTER => faster_fifo, 
		SLOWER => slower_fifo, 
		EMPTY => empty,
		FULL => full,
		CTRL_CLK => CTRL_CLK
		);

	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	--
	-- Lock-detection 
	--
	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

	i_lock_count: lock_count 
	GENERIC MAP(COUNT_LOCK => 6)
	PORT MAP(
		FASTER => faster_fifo,
		SLOWER => slower_fifo,
		EMPTY => empty,
		FULL => full,	
		CTRL_CLK => CTRL_CLK,
		CHECK_LOCK => rd_en,
		LOCK => lock_assumed
	);

	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	--
	-- State-machine for control of frequency-alignment
	--
	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	process(CTRL_CLK)
	begin
		if rising_edge(CTRL_CLK) then
			if RESET='1' then
				cstate <= IDLE;
				lock_r<='0';
			else
				cstate <= nstate;
				lock_r <= lock;
			end if;
		end if;
	end process;
	DLL_LOCKED <= lock_r;

	process(ctrl_clk)
	begin
		if rising_edge(ctrl_clk) then
			start_wait_fifo<='0';
			fifo_rst<='1';
			case cstate is
				when IDLE =>
					lock <= '0';
					fifo_rst <= '1';
					nstate <= EXACT;
				when EXACT =>
					fifo_rst<='0';
					start_wait_fifo <='1';
					lock <= '0';
					if lock_assumed = '1' then
						nstate <= LOCKED;
					else
						nstate <= EXACT;
					end if;
				when LOCKED =>
					lock <= '1';
					fifo_rst<='0';
					if STOP_LOCK_CHECK='1' then
						nstate<= LOCKED;
					else
						if (empty = '1') or (full = '1') then
							nstate <= EXACT;
						else
							nstate <= LOCKED;
						end if;
					end if;
				when others =>
					nstate<=IDLE;
			end case;
		end if;
	end process;

	process(DLL_CLK)
	begin
		if rising_edge(DLL_CLK) then
			if RESET = '1' then
				fifo_rst_reg <= '1';
			else
				fifo_rst_reg <= fifo_rst;
			end if;
		end if;
	end process;

	process(DLL_CLK)
	begin
		if rising_edge(DLL_CLK) then
			if (RESET = '1') then
				rd_en <= '0';
			elsif start_wait_fifo='1'  then
				rd_en <= '1';
			end if;
		end if;
	end process;

end Behavioral;

