--
--/////////////////////////////////////////////////////////////////////////////
--
--    File Name:  fifo_ctrl_r1.0.vhd
--      Version:  1.0
--         Date:  22.02.2008
--        Model:  
--
--      Company:  Xilinx, Inc.
--  Contributor:  
--
--   Disclaimer:  XILINX IS PROVIDING THIS DESIGN, CODE, OR
--                INFORMATION "AS IS" SOLELY FOR USE IN DEVELOPING
--                PROGRAMS AND SOLUTIONS FOR XILINX DEVICES.  BY
--                PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
--                ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
--                APPLICATION OR STANDARD, XILINX IS MAKING NO
--                REPRESENTATION THAT THIS IMPLEMENTATION IS FREE
--                FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE
--                RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY
--                REQUIRE FOR YOUR IMPLEMENTATION.  XILINX
--                EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH
--                RESPECT TO THE ADEQUACY OF THE IMPLEMENTATION,
--                INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
--                REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
--                FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES
--                OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
--                PURPOSE.
--
--                (c) Copyright 2008 Xilinx, Inc.
--                All rights reserved.
--
--/////////////////////////////////////////////////////////////////////////////
-- 
-- Summary:
--  This module is responsible to control the faster- and slower-commands to the 
--  oscillator, based on the percentage how full the "FIFO" is. 
--  The control is based on the structure of a FIFO, however no memory is used. 
--  If the FIFO-Ctrl detects, that it is filling up to much then the command to 
--  the oscillator is issued to speed up. If the FIFO is moving towards empty then
--  the oscillator receives the command to slow down.
--  When the DLL is locked typically only the middle-positions of level_int are 
--  active.
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.NUMERIC_STD.ALL;
use IEEE.STD_LOGIC_TEXTIO.ALL;

--library std;
--use std.textio.all;

library UNISIM;
use UNISIM.VComponents.all;

library work;
use work.pkg_dll_reco.all;

entity fifo_ctrl_v2 is
	Generic( WR_ADDR_WIDTH: integer range 1 to 15:=4;
					 DLL_MULTIPLY: integer range 1 to 256:=2
	);
		Port ( WR_CLK : in  STD_LOGIC;
					 WR_EN : in  STD_LOGIC;
					 RD_CLK : in  STD_LOGIC;
					 RD_EN : in  STD_LOGIC;
					 RST : in  STD_LOGIC;
					 FASTER : out  STD_LOGIC:='0';
					 SLOWER : out  STD_LOGIC:='0';
					 CTRL_CLK : in  STD_LOGIC;
					 FULL : out STD_LOGIC:='0';
					 EMPTY : out STD_LOGIC:='0'
					 );	
end fifo_ctrl_v2;

architecture Behavioral of fifo_ctrl_v2 is
	constant max_size_wr : integer := 2 ** (WR_ADDR_WIDTH + 1) - 1;
	constant max_size_rd : integer := ((max_size_wr + 1) * DLL_MULTIPLY)-1;
	constant max_size_rd_real : real :=real(max_size_rd);

	constant RD_ADDR_WIDTH:integer := round_up_func(log_bin_func(max_size_rd_real));
	--constant RD_ADDR_HALF:integer := max_size_rd / 2;
	constant RD_ADDR_HALF:integer := max_size_rd / 4;
	constant WR_ADDR_HALF:integer := RD_ADDR_HALF;
	constant WR_ADDR_FULL:integer := RD_ADDR_HALF * 2;

	constant WR_ADDR_HALF_SL : STD_LOGIC_VECTOR(RD_ADDR_WIDTH downto 0) := conv_std_logic_vector(WR_ADDR_HALF, RD_ADDR_WIDTH+1);

	constant OUTER_FAST_BOUNDARY : integer := max_size_rd * 3 / 16;
	constant INNER_FAST_BOUNDARY : integer := max_size_rd * 6 / 16; 
	constant NEUTRAL : integer := max_size_rd / 2;
	constant INNER_SLOW_BOUNDARY : integer := max_size_rd - INNER_FAST_BOUNDARY - 1;
	constant OUTER_SLOW_BOUNDARY : integer := max_size_rd - OUTER_FAST_BOUNDARY - 1;
	

	signal wr_addr : std_logic_vector(RD_ADDR_WIDTH downto 0) := conv_std_logic_vector(WR_ADDR_HALF, RD_ADDR_WIDTH+1);
	signal wr_addr_rd_c : std_logic_vector(RD_ADDR_WIDTH downto 0);
	signal wr_addr_rd : std_logic_vector(RD_ADDR_WIDTH downto 0) := conv_std_logic_vector(WR_ADDR_HALF, RD_ADDR_WIDTH+1);
	
	signal rd_addr : std_logic_vector(RD_ADDR_WIDTH downto 0) :=(others => '0');

	constant WR_ADDR_GRAY_RESET : std_logic_vector(RD_ADDR_WIDTH downto 0):= ('0' & WR_ADDR_HALF_SL (RD_ADDR_WIDTH downto 1)) xor WR_ADDR_HALF_SL ;
	signal wr_addr_gray : std_logic_vector(RD_ADDR_WIDTH downto 0) := WR_ADDR_GRAY_RESET;
	signal wr_addr_gray_rd : std_logic_vector(RD_ADDR_WIDTH downto 0) := WR_ADDR_GRAY_RESET;
	
	
	signal wr_addr_c : std_logic_vector(RD_ADDR_WIDTH downto 0);
	signal rd_addr_c : std_logic_vector(RD_ADDR_WIDTH downto 0);
	
	signal wr_addr_12_16 : std_logic_vector(RD_ADDR_WIDTH downto 0);
	signal wr_addr_9_16 : std_logic_vector(RD_ADDR_WIDTH downto 0);
	signal wr_addr_1_2 : std_logic_vector(RD_ADDR_WIDTH downto 0);
	signal wr_addr_6_16 : std_logic_vector(RD_ADDR_WIDTH downto 0);
	signal wr_addr_3_16 : std_logic_vector(RD_ADDR_WIDTH downto 0);
	
	signal full_int, empty_int : std_logic := '0';
	signal full_set, empty_set : std_logic := '0';
	signal full_ctlr, empty_ctlr : std_logic := '0';

	
	signal diff_int, diff_prev: std_logic_vector(3 downto 0):=(others=> '0');
	
	signal level_int: std_logic_vector(4 downto 0):="00111";
	signal level_int_prev: std_logic_vector(4 downto 0):="00111";
	--level_int = 00001:  3/16 full
	--level_int = 00011:  6/16 full
	--level_int = 00111:  1/2
	--level_int = 01111:  10/16
	--level_int = 11111:  13/16 full
	signal active_level: std_logic_vector(4 downto 0):="00111";
	signal level_toggle: std_logic;

	signal tendency, tendency_prev, tendency_prev2: std_logic_vector(1 downto 0):=(others => '0');
	signal resulting_tendency: std_logic_vector(1 downto 0):=(others => '0');
	
	signal faster_rdclk, slower_rdclk : std_logic := '0';

	signal faster_vec, slower_vec: std_logic_vector(2 downto 0):="000";
	
	signal set_faster, set_slower: std_logic;
	signal clear_faster, clear_slower: std_logic:='0';

	constant max_command_cnt : integer := 23 * DLL_MULTIPLY;
	signal command: std_logic:='0';
	signal command_cnt: integer range 0 to max_command_cnt := 0;
	signal command_cnt_tc: std_logic:='0';

	signal up_coarse, down_coarse: std_logic:='0';
	signal coarse_done: std_logic:='0';


----------------------

	constant max_cnt_coarse_wr: integer :=round_up_func(real(100/DLL_MULTIPLY));
	constant max_cnt_coarse_wr_width : integer := round_up_func(log_bin_func(real(max_cnt_coarse_wr)))+1;
	constant max_cnt_coarse_rd_width : integer := round_up_func(log_bin_func(real(DLL_MULTIPLY))) + max_cnt_coarse_wr_width;
	
	signal cnt_wr : std_logic_vector(max_cnt_coarse_wr_width downto 0):=(others=>'0');--driven by Refclk
	signal cnt_rd : std_logic_vector(max_cnt_coarse_rd_width downto 0):=(others=>'0');--driven by DLL-clock
	
	signal rst_cnt_rd: std_logic:='0';--\ synchronous reset for the two
	signal rst_cnt_wr: std_logic:='0';--/ counters
	
	signal cnt_rd_run: std_logic:='0';--\ enable-signals for the two
	signal cnt_wr_run: std_logic:='0';--/ counters
	signal cnt_rd_stopped: std_logic:='0';
	
	--===========================

	
	signal debug_dist: integer range -10000 to 10000;
begin
	--pragma translate_off
		--Display of calculated constants in Simulator
		process
			constant c:character :=CR;
		begin
			if now=0 ns then
				report c & c &			 
							 "Calculated Constants" & c &
							 "~~~~~~~~~~~~~~~~~~~~" & c &
							 "max_cnt_coarse_wr: "  & integer'image(max_cnt_coarse_wr)& c & 
							 "max_cnt_coarse_wr_width: "  & integer'image(max_cnt_coarse_wr_width)& c & 
							 "max_cnt_coarse_rd_width: "  & integer'image(max_cnt_coarse_rd_width)& c & 
							 "max_size_wr: " & integer'image(max_size_wr)& c &
							 "max_size_rd: " & integer'image(max_size_rd)& c &
							 --"max_size_actual_wr: " & integer'image(max_size_actual_wr)& c &
							 "RD_ADDR_WIDTH: " & integer'image(RD_ADDR_WIDTH)& c &
							 --"WR_ADDR_ACTUAL_WIDTH: " & integer'image(WR_ADDR_ACTUAL_WIDTH)& c &
							 "RD_ADDR_HALF: " & integer'image(RD_ADDR_HALF)& c &
							 "WR_ADDR_HALF: " & integer'image(WR_ADDR_HALF)& c &
							 "WR_ADDR_HALF_SL: " & integer'image(conv_integer(WR_ADDR_HALF_SL)) & c &
							 "WR_ADDR_FULL: " & integer'image(WR_ADDR_FULL)& c &
							 "WR_ADDR_GRAY_RESET:" & integer'image(conv_integer(WR_ADDR_GRAY_RESET))& c & 
							 "OUTER_FAST_BOUNDARY: " & integer'image(OUTER_FAST_BOUNDARY)& c &
							 "INNER_FAST_BOUNDARY: " & integer'image(INNER_FAST_BOUNDARY)& c &
							 "INNER_SLOW_BOUNDARY: " & integer'image(INNER_SLOW_BOUNDARY)& c &
							 "OUTER_SLOW_BOUNDARY: " & integer'image(OUTER_SLOW_BOUNDARY);
								
				wait;
			end if;
		end process;
	--pragma translate_on

	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	--
	-- Write-Counter and coding to Gray for clock-transfer
	--
	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	process(WR_CLK)
		
	begin
		if rising_edge(WR_CLK) then
			if RST = '1' or coarse_done='0' then
				wr_addr <= WR_ADDR_HALF_SL;
				wr_addr_gray <= WR_ADDR_GRAY_RESET; 			
			elsif WR_EN = '1' then
				wr_addr <= wr_addr_c;				
				wr_addr_gray <= ('0' & wr_addr_c(RD_ADDR_WIDTH downto 1)) xor wr_addr_c;
			end if;		
		end if;
	end process;
	wr_addr_c <= wr_addr  + DLL_MULTIPLY;
	

	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	--
	-- Read-Counter
	--
	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	process(RD_CLK)	
	begin
		if rising_edge(RD_CLK) then
			if RST = '1'  or coarse_done='0' then
				rd_addr <= (others => '0');
			elsif RD_EN='1' then
				rd_addr <= rd_addr_c;
			end if;
		end if;
	end process;
	rd_addr_c <= rd_addr + '1';
	
	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	--
	-- Clk-Domain transfer and re-coding to binary
	--
	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	process(RD_CLK)	
	begin
		if rising_edge(RD_CLK) then
			if RST = '1' then
				wr_addr_gray_rd <= WR_ADDR_GRAY_RESET;
				wr_addr_rd <= WR_ADDR_HALF_SL;
				-- Level-counters:
				wr_addr_12_16 <= WR_ADDR_HALF_SL + OUTER_SLOW_BOUNDARY; 
				wr_addr_9_16 <= WR_ADDR_HALF_SL + INNER_SLOW_BOUNDARY; 
				wr_addr_1_2 <= WR_ADDR_HALF_SL + NEUTRAL; 
				wr_addr_6_16 <= WR_ADDR_HALF_SL + INNER_FAST_BOUNDARY; 
				wr_addr_3_16 <= WR_ADDR_HALF_SL + OUTER_SLOW_BOUNDARY; 

			else
				wr_addr_gray_rd <= wr_addr_gray;
				wr_addr_rd <= wr_addr_rd_c;
				--Level counters:
				wr_addr_12_16 <= wr_addr_rd_c + OUTER_SLOW_BOUNDARY;
				wr_addr_9_16 <= wr_addr_rd_c + INNER_SLOW_BOUNDARY;
				wr_addr_1_2 <= wr_addr_rd_c + NEUTRAL;
				wr_addr_6_16 <= wr_addr_rd_c + INNER_FAST_BOUNDARY;
				wr_addr_3_16 <= wr_addr_rd_c + OUTER_FAST_BOUNDARY;
			end if;
		end if;
	end process;
	wr_addr_rd_c <= wr_addr_gray_rd xor ('0' & wr_addr_rd_c(RD_ADDR_WIDTH downto 1));
	
	
	debug_dist <= (conv_integer(wr_addr_rd) - conv_integer(rd_addr)) when (conv_integer(wr_addr_rd) > conv_integer(rd_addr)) else
								 max_size_rd - (conv_integer(rd_addr)-conv_integer(wr_addr_rd));
	
	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	--
	-- Generation of Full- and Empty-Flag
	--
	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	process(RD_CLK)	
	begin
		if rising_edge(RD_CLK) then
			if RST = '1' then
				full_int <= '0';
				empty_int <= '0';
			else
				if wr_addr_rd= rd_addr then
					empty_int <= '1';
				else
					empty_int <= '0';
				end if;
				if (wr_addr_rd(RD_ADDR_WIDTH) /= rd_addr(RD_ADDR_WIDTH)) and
					 (wr_addr_rd(RD_ADDR_WIDTH-1) /= rd_addr(RD_ADDR_WIDTH-1)) and
					 (wr_addr_rd(RD_ADDR_WIDTH - 2 downto 0) = rd_addr(RD_ADDR_WIDTH - 2 downto 0)) then
					full_int <= '1';
				else
					full_int <= '0';
				end if;
			end if;
		end if;
	end process;

	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	--
	-- Transfer of Full- and Empty-Flag into CTRL_CLK-domain
	--
	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

	process(RD_CLK)
	begin
		if rising_edge(RD_CLK) then
			if RST = '1' then
				full_set <= '0';
				empty_set <= '0';
			else
				if full_int = '1' then
					full_set <= '1';
				elsif full_ctlr ='1' then
					full_set <= '0';
				end if;
				if empty_int = '1' then
					empty_set <= '1';
				elsif empty_ctlr ='1' then
					empty_set <='0';
				end if;
			end if;
		end if;
	end process;
	
	process(CTRL_CLK)	
	begin
		if rising_edge(CTRL_CLK) then
			full_ctlr <= full_set;
			empty_ctlr <= empty_set;
		end if;
	end process;
			
				
	FULL <= full_ctlr;
	EMPTY <= empty_ctlr;
	

	
	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	--
	-- Check on level of FIFO-filling
	--
	--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
	process(RD_CLK)	
	begin
		if rising_edge(RD_CLK) then
			if RST = '1' then
				level_int <= "00111";
				level_int_prev <= "00111";
			else
				level_int_prev <= level_int;
				--12/16-FULL: (=outer slow boundary)
				if ((wr_addr_12_16(RD_ADDR_WIDTH-1) /= rd_addr(RD_ADDR_WIDTH-1)) and
					 (wr_addr_12_16(RD_ADDR_WIDTH - 2 downto 0)>= rd_addr(RD_ADDR_WIDTH - 2 downto 0))) or
					 ((wr_addr_12_16(RD_ADDR_WIDTH-1) = rd_addr(RD_ADDR_WIDTH-1)) and
					 (wr_addr_12_16(RD_ADDR_WIDTH - 2 downto 0) < rd_addr(RD_ADDR_WIDTH - 2 downto 0))) then
					level_int(4) <= '1';
				else
					level_int(4) <= '0';
				end if;

				--9/16-FULL: (=inner slow boundary)
				if ((wr_addr_9_16(RD_ADDR_WIDTH-1) /= rd_addr(RD_ADDR_WIDTH-1)) and
					 (wr_addr_9_16(RD_ADDR_WIDTH - 2 downto 0)>= rd_addr(RD_ADDR_WIDTH - 2 downto 0))) or
					 ((wr_addr_9_16(RD_ADDR_WIDTH-1) = rd_addr(RD_ADDR_WIDTH-1)) and
					 (wr_addr_9_16(RD_ADDR_WIDTH - 2 downto 0) < rd_addr(RD_ADDR_WIDTH - 2 downto 0))) then
					level_int(3) <= '1';
				else
					level_int(3) <= '0';
				end if;

				--HALF-FULL: (=neutral area)
				if ((wr_addr_1_2(RD_ADDR_WIDTH-1) /= rd_addr(RD_ADDR_WIDTH-1)) and
					 (wr_addr_1_2(RD_ADDR_WIDTH - 2 downto 0)>= rd_addr(RD_ADDR_WIDTH - 2 downto 0))) or
					 ((wr_addr_1_2(RD_ADDR_WIDTH-1) = rd_addr(RD_ADDR_WIDTH-1)) and
					 (wr_addr_1_2(RD_ADDR_WIDTH - 2 downto 0) < rd_addr(RD_ADDR_WIDTH - 2 downto 0))) then
					level_int(2) <= '1';
				else
					level_int(2) <= '0';
				end if;
				
				--6/16 FULL (=inner fast boundary)
				if ((wr_addr_6_16(RD_ADDR_WIDTH-1) /= rd_addr(RD_ADDR_WIDTH-1)) and
					 (wr_addr_6_16(RD_ADDR_WIDTH - 2 downto 0)>= rd_addr(RD_ADDR_WIDTH - 2 downto 0))) or
					 ((wr_addr_6_16(RD_ADDR_WIDTH-1) = rd_addr(RD_ADDR_WIDTH-1)) and
					 (wr_addr_6_16(RD_ADDR_WIDTH - 2 downto 0) < rd_addr(RD_ADDR_WIDTH - 2 downto 0))) then-- or
					level_int(1) <= '1';
				else
					level_int(1) <= '0';
				end if;
				
				--3/16 FULL (=outer fast boundary)
				if ((wr_addr_3_16(RD_ADDR_WIDTH-1) /= rd_addr(RD_ADDR_WIDTH-1)) and
					 (wr_addr_3_16(RD_ADDR_WIDTH - 2 downto 0)>= rd_addr(RD_ADDR_WIDTH - 2 downto 0))) or
					 ((wr_addr_3_16(RD_ADDR_WIDTH-1) = rd_addr(RD_ADDR_WIDTH-1)) and
					 (wr_addr_3_16(RD_ADDR_WIDTH - 2 downto 0) < rd_addr(RD_ADDR_WIDTH - 2 downto 0))) then
					level_int(0) <= '1';
				else
					level_int(0) <= '0';
				end if;
			end if;
		end if;
	end process;

        level_toggle <= '1' when level_int_prev /= level_int else '0';

	process(RD_CLK)
	begin
		if rising_edge(RD_CLK) then
		--	if level_int_prev /= level_int then
			if level_toggle = '1' then
				tendency<="00";
				slow_check: for i in 0 to 3 loop
					if (level_int(i)='1' and level_int_prev(i)='0' and level_int(i+1)='1' and level_int_prev(i+1)='1') or
						 (level_int(i)='1' and level_int_prev(i)='1' and level_int(i+1)='0' and level_int_prev(i+1)='1') then
						tendency(1) <= '1';
					end if;
				end loop;

				fast_check: for i in 0 to 3 loop
					if (level_int(i) = '1' and level_int_prev(i)='1' and level_int(i+1)='1' and level_int_prev(i+1)='0') or
						 (level_int(i) = '0' and level_int_prev(i)='0' and level_int(i+1)='0' and level_int_prev(i+1)='1') or
						 (level_int(i) = '0' and level_int_prev(i)='1' and level_int(i+1)='1' and level_int_prev(i+1)='1')then
						tendency(0) <= '1';
					end if;
				end loop;

			end if;
				tendency_prev <= tendency;
				tendency_prev2 <= tendency_prev;

				resulting_tendency <= tendency and tendency_prev and tendency_prev2;

		end if;
	end process;


--Several situations can occur:
-- 1) LOCKED-sitation: FIFO moving around the half-full point
-- 2) RD-clock faster than WR-clock: => FIFO underflowing (during locking-phase)
-- 3) RD-clock slower than WR-clock: => FIFO overflowing (during locking-phase)
--    The FIFO is started with nominally half-full FIFO-pointers. If nothing would
--    be compensated (=RD-clock always slower) then first the FULL-flag is asserted
--    and the FIFO-pointers roll over. Later on, with the RD-clock assumed constant
--    the EMPTY-flag is asserted and again the pointers start catching up leading
--    to a FULL-condition (and so on).
--
--    If it is assumed that during the locking-phase over- and underflows are
--    allowed then care must be taken to distinguish between the faster and the
--    slower read-situation (with the write-clock being constant).
--    A (constantly) slower read-clock creates a FULL-flag, overflows and then goes
--    on until until it is overtaken by the write-pointer, leading to an EMPTY-flag
--    being asserted.
--    Even though after LOCK both over- or underflowing are not allowed, they can
--    be used to get into the stable middle-condition:
--    If a overflow has happended, it is known that the readpointer was to slow, the
--    speed of reading needs to be increased.
--    As the symptoms for underflow (= (Empty = 1) & (level_int = 11000)) or overflow
--    (= (Full=1) & (level_int = 00111)) are the same for both a faster and a slower
--    read-clock, it is necessary to have a tendency-information.
--    If the 1s in level_int are shifting from 4=>0 then the read-clock is too slow,
--    on a shift of 1s from 0=>4 the read-clock is too fast. As the tendancy is only
--    happening relatively slowly, it can only be calculated on changes of the
--    level_int-signal.
--    For the overflow/underflow this means that the full/empty-signal is taken and
--    checked with the tendency-signal. If the full/empty-signal is asserted with a
--    tendency of a read-clock being faster than the write-clock then the tap-number
--    must be increased => DLL needs to be slower.
--    If the full/empty-signal is asserted with a tendency of the read-clock being
--    slower than the write-clock, then the read-clock needs to be speeded up by
--    decreasing the number of taps in the DLL.

	process(RD_CLK)
	begin
		if rising_edge(RD_CLK) then
			if RST = '1' then
				up_coarse <= '0';
				down_coarse <= '0';
				coarse_done <= '0';
			else
				if slower_rdclk = '1' then
					down_coarse <= '1';
				end if;

				if faster_rdclk = '1' then
					up_coarse <= '1';
				end if;

				if up_coarse = '1' and down_coarse ='1' then
					coarse_done <= '1';
				end if;
			end if;
		end if;
	end process;

-- Following Level_int-combination have a proper meaning:
-- Level_int(4..0)   FIFO-filling:
--   11000              very close to empty => go slower
--   11100              close to empty => too fast => go slower
--   11110              just below half-full => too fast => go slower
--   01110              half-full (about 1/4th in the middle) ==> wanted situation
--   01111              just above half-full => too slow => go faster
--   00111              close to full => too slow => go faster
--   00011              very close to full, go faster


	process(RD_CLK)
		variable level_int_v: std_logic_vector(4 downto 0);
	begin
		if rising_edge(RD_CLK) then
			if RST = '1' then
				faster_rdclk <= '0';
				slower_rdclk <= '0';
				--active_level <= (others=>'0');
				active_level <= "11100";
			elsif RD_EN = '1' then
				faster_rdclk <= '0';
				slower_rdclk <= '0';
				--two situations need to be considered:
				-- a) before lock, over/underflows can occur
				-- b) after lock, FIFO is moving within its valid range
				--

				if level_int /= level_int_prev then
					level_int_v := level_int;
				else
					level_int_v :=(others=>'0');
				end if;

				if command = '0' then

					if coarse_done = '0' then
					-- a) Empty-Check for time before LOCK

						--Two counters running against each other
						if cnt_rd_stopped='1' and rst_cnt_rd='0' then
							if cnt_rd > (conv_integer(cnt_wr) * DLL_MULTIPLY) then
								slower_rdclk <= '1';
							elsif cnt_wr /= 0 then--Checking if this is running and > 0
								faster_rdclk <='1';
							end if;
						end if;
							if cnt_rd_stopped = '1' then--handshaking for restart
								rst_cnt_rd <= '1';
							else
								rst_cnt_rd <= '0';
							end if;
					else
					-- b) Check within the valid range after lock
					--
					-- Only level-changes that are coming from the inner range have an effect on
					-- the number of taps.
						case level_int_v is
							when "00000" =>
								null;
							when "11000" =>
								active_level <= "11000";
								if active_level /= "11000" then
									slower_rdclk <= '1';
								end if;
							when "11100" =>
								active_level <= "11100";
								if (active_level /= "11100") and (active_level /= "11000") then
									slower_rdclk <= '1';
								end if;
							when "11110" =>
								active_level <= "11110";
								if (active_level /= "11110") and (active_level /= "11100") and (active_level /= "11000") then
									slower_rdclk <= '1';
								end if;
							when "01110" =>
								null;
							when "01111" =>
								active_level <= "01111";
								if (active_level /= "01111") and (active_level /= "00111") and (active_level /= "00011") then
									faster_rdclk <= '1';
								end if;
							when "00111" =>
								active_level <= "00111";
								if (active_level /= "00111") and (active_level /= "00011") then
									faster_rdclk <= '1';
								end if;
							when "00011" =>
								active_level <= "00011";
								if active_level /= "00011" then
									faster_rdclk <= '1';
								end if;
							when others => 
								null;
						end case;
					end if;
				end if;
			end if;
		end if;
	end process;

	--##################################################
	-- 
	-- Coarse-Loop Counters
	--
	--##################################################
	
	process(RD_CLK)--DLL-Clock
	begin
		if rising_edge(RD_CLK) then
			if rst_cnt_rd ='1' then
				cnt_rd <=(others=>'0');
			elsif coarse_done='0' and cnt_rd_run='1' then
				cnt_rd <= cnt_rd + 1;
			end if;
			cnt_rd_stopped <= not cnt_rd_run; --one
		end if;
	end process;


	
	process(RD_CLK)-- = DLL-Clock
	begin
		if rising_edge(RD_CLK) then
			cnt_rd_run <= cnt_wr_run; --resynchronisation to Rd_CLK
		end if;
	end process;

	process(WR_CLK)-- = Ref-Clock
	begin
		if rising_edge(WR_CLK) then
			rst_cnt_wr <= rst_cnt_rd;
		end if;
	end process;


	process(WR_CLK)-- = Ref-Clock
	begin
		if rising_edge(WR_CLK) then
			if rst_cnt_wr ='1' then
				cnt_wr <=(others=>'0');		
			elsif coarse_done='0' and cnt_wr_run='1' then
				cnt_wr <= cnt_wr + 1;
			end if;
			if cnt_wr = max_cnt_coarse_wr then
				cnt_wr_run <= '0';
			elsif rst_cnt_wr='1' then
				cnt_wr_run <='1';
			end if;
		end if;
	end process;


	--##################################################
	--
	-- Damping to avoid too fast switching
	--
	--##################################################

	process(RD_CLK)
	begin
		if rising_edge(RD_CLK) then
			if RST = '1' then
				command <='0';
			elsif (coarse_done = '1') and (faster_rdclk = '1' or slower_rdclk = '1') then
				command <= '1';
			elsif command_cnt_tc = '1' and level_toggle ='0' then
				command <='0';
			end if;
		end if;
	end process;

	process(CTRL_CLK)
	begin
		if rising_edge(CTRL_CLK) then
			if RST = '1' then
				command_cnt <= 0;
			elsif command = '1' then
				command_cnt <= command_cnt + 1;
				if command_cnt >= max_command_cnt - 3 then
					command_cnt_tc <= '1';
				else
					command_cnt_tc <= '0';
				end if;
			else
				command_cnt <= 0;
				command_cnt_tc <= '0';
			end if;
		end if;
	end process;


	--##################################################
	--
	-- Translation into CTRLCLK-domain
	--
	--##################################################



	faster_ff:fdc port map(
		D => '1',
		C => faster_rdclk,
		CLR=> clear_faster,
		Q => set_faster);

	slower_ff:fdc port map(
		D => '1',
		C => slower_rdclk,
		CLR => clear_slower,
		Q => set_slower);

	faster_vec(2) <= set_faster;
	slower_vec(2) <= set_slower;

	process(CTRL_CLK)
	begin
		if rising_edge(CTRL_CLK) then
			if RST = '1' then
				faster_vec(1 downto 0) <= "00";
				slower_vec(1 downto 0) <= "00";
				FASTER <= '0';
				SLOWER <= '0';
				clear_faster <= '0';
				clear_slower <= '0';
			else
				if RD_EN = '1' then

					faster_vec(1 downto 0) <= faster_vec(2 downto 1);
					slower_vec(1 downto 0) <= slower_vec(2 downto 1);
					if faster_vec = "110" then
						clear_faster <= '1';
						FASTER <= '1';
					else
						clear_faster <= '0';
						FASTER <= '0';
					end if;
					if slower_vec = "110" then
						SLOWER <= '1';
						clear_slower <= '1';
					else
						clear_slower <= '0';
						SLOWER <= '0';
					end if;
				else
					FASTER <= '0';
					SLOWER <= '0';
				end if;
			end if;
		end if;
	end process;

end Behavioral;
